#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "basic.h"
#include "treiber.h"

#define CAS(x,y,z) __sync_bool_compare_and_swap(x,y,z)

static int dbg = 0 ;

/* Nodes */

static stack_node_t *alloc_stack_node(void *val) {
  stack_node_t *r = malloc_check(sizeof(*r)) ;
  r->val = val ; r->next = NULL ;
  return r ;
}

static void free_stack_node(stack_node_t *p) { free(p) ; }

stack_t *alloc_stack(int nprocs) {
  stack_t *r =  malloc_check(sizeof(*r)) ;
  r->top = NULL ;
  return r ;
}

void free_stack(stack_t *p) { free(p) ; }

void *pop(stack_t *p,int id) {
  for ( ; ; ) {
    SYNC() ;
    stack_node_t *old = p->top ; // Old top node.
    SYNC() ;
    if (old == NULL) {
      return NULL ;
    } else {
      stack_node_t *next = old->next ;
      SYNC() ;
      if (CAS(&p->top,old,next)) { // Do pop
        SYNC() ;
        if (dbg) fprintf(stderr,"POP: %p\n",old) ;
        void *r = old->val ;         // Clean up and return
        free_stack_node(old) ;
        return r ;
      }
    }
  }
}


void push(stack_t *p, void *v) {
  stack_node_t *q = alloc_stack_node(v), *old ;
  do {
    SYNC() ;
    old = p->top ;
    SYNC() ;
    q->next = old ;                        
    SYNC() ;
  } while (!CAS(&p->top,old,q)) ; // Do push
  SYNC() ;
  if (dbg) fprintf(stderr,"PUSH: %p\n",q) ;
}
