#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "fifo.h"

static int verbose = 0 ;
static int n = 2 ;
static int m = 1 ;



/* Barrier */

typedef struct {
  unsigned int nprocs ;
  volatile unsigned int count ;
  pthread_mutex_t *mutex ;
  pthread_cond_t *cond ;
} barrier_t ;

barrier_t *alloc_my_barrier(unsigned int nprocs) {
  barrier_t *p = malloc_check(sizeof(*p)) ;
  p->nprocs = p->count = nprocs ;
  p->mutex = alloc_mutex() ;
  p->cond = alloc_cond() ;
  return p ;
}

void free_my_barrier(barrier_t *p) {
  free_cond(p->cond) ;
  free_mutex(p->mutex) ;
  free(p) ;
}

int wait_my_barrier(barrier_t *p) {
  int ret ;
  lock_mutex(p->mutex) ;
  --p->count ;
  if (p->count > 0) {
    wait_cond(p->cond,p->mutex) ; ret = 0 ;
  } else {
    broadcast_cond(p->cond) ;
    p->count = p->nprocs ;
    ret = PTHREAD_BARRIER_SERIAL_THREAD ;
  }
  unlock_mutex(p->mutex) ;
  return ret ;
}

typedef struct {
  barrier_t *b ;  
  int m ;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

void *runner(void *_p) {
  ctx_t *p = _p ; common_t *q = p->common ;
  for (int k = 0 ; k < q->m ; k++) {
    wait_my_barrier(q->b) ;
    printf("<%i>",p->id) ; fflush(stdout) ;
    delay(0) ;
    (void)wait_my_barrier(q->b) ;
    if (p->id == 0)  printf("\n") ;
  }
  return NULL ;
}

void run(int n, int m) {
  common_t c ;
  c.b = alloc_my_barrier(n) ; c.m = m ;

  pthread_t th[n] ; ctx_t a[n] ;
  for (int k = 0 ; k < n ; k++) {
    ctx_t *p = &a[k] ; p->id = k ; p->common = &c ;
    create_thread(&th[k],runner,p) ;
  }

  for (int k = 0 ; k < n ; k++) (void)join_thread(&th[k]) ;
  free_my_barrier(c.b) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n] [m]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  while (argc > 1) {
    --argc ; ++argv ;
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    
  }
  if (argc > 1) {
    n = atoi(argv[0]) ;
    m = atoi(argv[1]) ;
  } else if (argc > 0) {
    n = atoi(argv[0]) ;
  }  
  run(n,m) ;
  return 0 ;
}
